package caps

import (
	"errors"
	"sort"
)

type (
	UserCapability   string
	UserCapabilities map[UserCapability]int
)

const (
	USERCAP_USER_CREATE   UserCapability = "USER_CREATE"
	USERCAP_USER_DESTROY  UserCapability = "USER_DESTROY"
	USERCAP_USER_SET_CAPS UserCapability = "USER_SET_CAPS"
	USERCAP_USER_UNLOCK   UserCapability = "USER_UNLOCK"
	USERCAP_USER_RESETPWD UserCapability = "USER_RESETPWD"
	USERCAP_USER_LIST     UserCapability = "USER_LIST"

	USERCAP_VM_ATTACH_ANY     UserCapability = "VM_ATTACH_ANY"
	USERCAP_VM_START_ANY      UserCapability = "VM_START_ANY"
	USERCAP_VM_STOP_ANY       UserCapability = "VM_STOP_ANY"
	USERCAP_VM_REBOOT_ANY     UserCapability = "VM_REBOOT_ANY"
	USERCAP_VM_CREATE         UserCapability = "VM_CREATE"
	USERCAP_VM_DESTROY_ANY    UserCapability = "VM_DESTROY_ANY"
	USERCAP_VM_EDIT_ANY       UserCapability = "VM_EDIT_ANY"
	USERCAP_VM_SET_ACCESS     UserCapability = "VM_SET_ACCESS"
	USERCAP_VM_SET_ACCESS_ANY UserCapability = "VM_SET_ACCESS_ANY"
	USERCAP_VM_READFS_ANY     UserCapability = "VM_READFS_ANY"
	USERCAP_VM_WRITEFS_ANY    UserCapability = "VM_WRITEFS_ANY"

	USERCAP_EXAM_ATTACH UserCapability = "EXAM_ATTACH"

	USERCAP_TPL_CREATE      UserCapability = "TPL_CREATE"
	USERCAP_TPL_EDIT        UserCapability = "TPL_EDIT"
	USERCAP_TPL_EDIT_ANY    UserCapability = "TPL_EDIT_ANY"
	USERCAP_TPL_LIST        UserCapability = "TPL_LIST"
	USERCAP_TPL_LIST_ANY    UserCapability = "TPL_LIST_ANY"
	USERCAP_TPL_DESTROY     UserCapability = "TPL_DESTROY"
	USERCAP_TPL_DESTROY_ANY UserCapability = "TPL_DESTROY_ANY"
	USERCAP_TPL_READFS      UserCapability = "TPL_READFS"
	USERCAP_TPL_READFS_ANY  UserCapability = "TPL_READFS_ANY"

	_USERCAP_INVALID_ UserCapability = "_USERCAP_INVALID_"
)

var userCapsList = []UserCapability{
	USERCAP_USER_CREATE,
	USERCAP_USER_DESTROY,
	USERCAP_USER_SET_CAPS,
	USERCAP_USER_UNLOCK,
	USERCAP_USER_RESETPWD,
	USERCAP_USER_LIST,

	USERCAP_VM_ATTACH_ANY,
	USERCAP_VM_START_ANY,
	USERCAP_VM_STOP_ANY,
	USERCAP_VM_REBOOT_ANY,
	USERCAP_VM_CREATE,
	USERCAP_VM_DESTROY_ANY,
	USERCAP_VM_EDIT_ANY,
	USERCAP_VM_SET_ACCESS,
	USERCAP_VM_SET_ACCESS_ANY,
	USERCAP_VM_READFS_ANY,
	USERCAP_VM_WRITEFS_ANY,

	USERCAP_EXAM_ATTACH,

	USERCAP_TPL_CREATE,
	USERCAP_TPL_EDIT,
	USERCAP_TPL_EDIT_ANY,
	USERCAP_TPL_LIST,
	USERCAP_TPL_LIST_ANY,
	USERCAP_TPL_DESTROY,
	USERCAP_TPL_DESTROY_ANY,
	USERCAP_TPL_READFS,
	USERCAP_TPL_READFS_ANY,
}

// Returns false if the string is invalid.
func NewUserCapability(s string) (UserCapability, bool) {
	switch s {
	case "USER_CREATE":
		return USERCAP_USER_CREATE, true
	case "USER_DESTROY":
		return USERCAP_USER_DESTROY, true
	case "USER_SET_CAPS":
		return USERCAP_USER_SET_CAPS, true
	case "USER_UNLOCK":
		return USERCAP_USER_UNLOCK, true
	case "USER_RESETPWD":
		return USERCAP_USER_RESETPWD, true
	case "USER_LIST":
		return USERCAP_USER_LIST, true

	case "VM_ATTACH_ANY":
		return USERCAP_VM_ATTACH_ANY, true
	case "VM_START_ANY":
		return USERCAP_VM_START_ANY, true
	case "VM_STOP_ANY":
		return USERCAP_VM_STOP_ANY, true
	case "VM_REBOOT_ANY":
		return USERCAP_VM_REBOOT_ANY, true
	case "VM_CREATE":
		return USERCAP_VM_CREATE, true
	case "VM_DESTROY_ANY":
		return USERCAP_VM_DESTROY_ANY, true
	case "VM_EDIT_ANY":
		return USERCAP_VM_EDIT_ANY, true
	case "VM_SET_ACCESS":
		return USERCAP_VM_SET_ACCESS, true
	case "VM_SET_ACCESS_ANY":
		return USERCAP_VM_SET_ACCESS_ANY, true
	case "VM_READFS_ANY":
		return USERCAP_VM_READFS_ANY, true
	case "VM_WRITEFS_ANY":
		return USERCAP_VM_WRITEFS_ANY, true

	case "EXAM_ATTACH":
		return USERCAP_EXAM_ATTACH, true

	case "TPL_CREATE":
		return USERCAP_TPL_CREATE, true
	case "TPL_EDIT":
		return USERCAP_TPL_EDIT, true
	case "TPL_EDIT_ANY":
		return USERCAP_TPL_EDIT_ANY, true
	case "TPL_LIST":
		return USERCAP_TPL_LIST, true
	case "TPL_LIST_ANY":
		return USERCAP_TPL_LIST_ANY, true
	case "TPL_DESTROY":
		return USERCAP_TPL_DESTROY, true
	case "TPL_DESTROY_ANY":
		return USERCAP_TPL_DESTROY_ANY, true
	case "TPL_READFS":
		return USERCAP_TPL_READFS, true
	case "TPL_READFS_ANY":
		return USERCAP_TPL_READFS_ANY, true

	default:
		return "", false
	}
}

func (cap UserCapability) String() string {
	return string(cap)
}

// Returns true if specified VMAccessCapability exists
func (cap UserCapability) IsValid() bool {
	return cap < _USERCAP_INVALID_
}

func (caps UserCapabilities) Add(capName string) error {
	cap, success := NewUserCapability(capName)
	if !success {
		return errors.New("invalid user capability")
	}
	caps[cap] = 1
	return nil
}

// Validates all user capabilities.
func (caps UserCapabilities) Validate() error {
	for cap := range caps {
		if !cap.IsValid() {
			return errors.New("invalid user capability: " + cap.String())
		}
	}
	return nil
}

// Returns true if user has at least one USERCAP_VM_xxx_ANY capability
func (caps UserCapabilities) HasVMxxxAnyCapability() bool {
	count := 0
	count += caps[USERCAP_VM_DESTROY_ANY]
	count += caps[USERCAP_VM_EDIT_ANY]
	count += caps[USERCAP_VM_START_ANY]
	count += caps[USERCAP_VM_STOP_ANY]
	count += caps[USERCAP_VM_REBOOT_ANY]
	count += caps[USERCAP_VM_ATTACH_ANY]
	count += caps[USERCAP_VM_READFS_ANY]
	count += caps[USERCAP_VM_WRITEFS_ANY]
	count += caps[USERCAP_VM_SET_ACCESS_ANY]
	return count > 0
}

func GetUserCapsNames() []string {
	sorted := make([]string, len(userCapsList))
	for i, cap := range userCapsList {
		sorted[i] = cap.String()
	}
	// Sort the sorted array in place
	sort.Strings(sorted)
	return sorted
}

//----------------------------------------------------------------------------------

type (
	VMAccessCapability   string
	VMAccessCapabilities map[VMAccessCapability]int
)

const (
	VMCAP_SET_ACCESS VMAccessCapability = "VM_SET_ACCESS"
	VMCAP_DESTROY    VMAccessCapability = "VM_DESTROY"
	VMCAP_EDIT       VMAccessCapability = "VM_EDIT"
	VMCAP_START      VMAccessCapability = "VM_START"
	VMCAP_STOP       VMAccessCapability = "VM_STOP"
	VMCAP_REBOOT     VMAccessCapability = "VM_REBOOT"
	VMCAP_ATTACH     VMAccessCapability = "VM_ATTACH"
	VMCAP_READFS     VMAccessCapability = "VM_READFS"
	VMCAP_WRITEFS    VMAccessCapability = "VM_WRITEFS"

	_VMCAP_INVALID_ = "_VMCAP_INVALID_"
)

var vmAccessCapsList = []VMAccessCapability{
	VMCAP_SET_ACCESS,
	VMCAP_DESTROY,
	VMCAP_EDIT,
	VMCAP_START,
	VMCAP_STOP,
	VMCAP_REBOOT,
	VMCAP_ATTACH,
	VMCAP_READFS,
	VMCAP_WRITEFS,
}

type VMActions struct {
	Attch    bool // attach
	Start    bool
	Stop     bool // shutdown/kill
	Reboot   bool
	Edit     bool
	Del      bool // delete
	ExpDir   bool // export dirs/files
	ImpDir   bool // import dirs/files
	VMAccess bool // change VM access
}

// Returns false if the string is invalid.
func NewVMAccessCapability(s string) (VMAccessCapability, bool) {
	switch s {
	case "VM_SET_ACCESS":
		return VMCAP_SET_ACCESS, true
	case "VM_DESTROY":
		return VMCAP_DESTROY, true
	case "VM_EDIT":
		return VMCAP_EDIT, true
	case "VM_START":
		return VMCAP_START, true
	case "VM_STOP":
		return VMCAP_STOP, true
	case "VM_REBOOT":
		return VMCAP_REBOOT, true
	case "VM_ATTACH":
		return VMCAP_ATTACH, true
	case "VM_READFS":
		return VMCAP_READFS, true
	case "VM_WRITEFS":
		return VMCAP_WRITEFS, true
	}
	return "", false
}

func (cap VMAccessCapability) String() string {
	return string(cap)
}

// Returns true if specified VMAccessCapability exists
func (cap VMAccessCapability) IsValid() bool {
	return cap < _VMCAP_INVALID_
}

func (caps VMAccessCapabilities) Add(capName string) error {
	cap, success := NewVMAccessCapability(capName)
	if !success {
		return errors.New("invalid VM access capability")
	}
	caps[cap] = 1
	return nil
}

// Validates all VM access capabilities.
func (caps VMAccessCapabilities) Validate() error {
	for cap := range caps {
		if !cap.IsValid() {
			return errors.New("invalid VM access capability: " + cap.String())
		}
	}
	return nil
}

func (caps VMAccessCapabilities) ToVMActions(isOwner bool) VMActions {
	actions := VMActions{}
	actions.Start = (caps[VMCAP_START] == 1 || isOwner)
	actions.Stop = (caps[VMCAP_STOP] == 1 || isOwner)
	actions.Del = (caps[VMCAP_DESTROY] == 1 || isOwner)
	actions.Reboot = (caps[VMCAP_REBOOT] == 1 || isOwner)
	actions.Edit = (caps[VMCAP_EDIT] == 1 || isOwner)
	actions.Attch = (caps[VMCAP_ATTACH] == 1 || isOwner)
	actions.ExpDir = (caps[VMCAP_READFS] == 1 || isOwner)
	actions.ImpDir = (caps[VMCAP_WRITEFS] == 1 || isOwner)
	actions.VMAccess = (caps[VMCAP_SET_ACCESS] == 1 || isOwner)
	return actions
}

func GetVMAccessCapsNames() []VMAccessCapability {
	return vmAccessCapsList
}
